from .adom import Document, Element, createTextElement, Action

elementTypes = [
	"cont",
	"text",
	"link",
	"input"
]

validFormats = [
	"m100"
]

def _crashDoc(reason: str, lineNumber: int):
	doc = Document(None)
	doc.elements = [createTextElement("Line {}: ".format(lineNumber + 1) + reason)]
	return doc

def term2doc(contents, browser):
	res = Document(browser)

	lines: list = contents.splitlines()
	
	elementQueue: list = []

	readTo = -1
	for lineNumber in range(len(lines)):
		command: str = lines[lineNumber].strip()

		if lineNumber == 0:
			if not command.startswith("@termtype"):
				return _crashDoc("Term files must start with @termtype:[a valid termtype].\nex. @termtype:m100", lineNumber)
			else:
				termstart = digestAttribute(command[1:])
				passes = False
				for tt in validFormats:
					if termstart["value"] == tt:
						passes = True
						break
				if not passes:
					return _crashDoc("Term file is not in a valid format.", lineNumber)
				continue
				
		if readTo > lineNumber:
			continue

		top: Element = getTop(elementQueue)

		if command.startswith("#"):
			continue
		elif command.startswith("-"):
			if top == None:
				return _crashDoc("No element for attribute to be assigned.", lineNumber)
			else:
				result: dict = digestAttribute(command[1:])

				if result == None:
					return _crashDoc("Could not interpret attribute.", lineNumber)
				else:
					top.setAttribute(result["name"], result["value"])
			
		elif command.startswith(tuple(elementTypes)):
			etype, evalue = digestDeclaration(command)
			newElement = Element(etype)
			newElement.value = evalue

			if top != None:
				top.children.append(newElement)

			if etype == "input":
				res.hasInputs = True
			
			elementQueue.append(newElement)
			
		elif command.startswith("end"):
			if top == None:
				return _crashDoc("No element for `end` keyword to complete.", lineNumber)
			else:
				ele: Element = elementQueue.pop()

				if ele.type == "link":
					linkKey = ele.getAttribute("key")
					linkURL = ele.getAttribute("url")
					if(linkKey == None or linkURL == None):
						return _crashDoc("Link element requires attributes (key, url)")
					res.add_link(linkKey, linkURL)

				if len(elementQueue) == 0:
					res.elements.append(ele)
		elif command.startswith("action"):
			comIndex = command.find(":")
			if comIndex == -1:
				return _crashDoc("Could not find `:` following the `action` declaration. This needs to be on the same line.", lineNumber)
			between = command[len("action"):comIndex]
			if between.strip() != "":
				return _crashDoc("Unrecognized `" + between + "`.", lineNumber)
			
			# determine action name
			startIndex = command.find("(", comIndex)
			if startIndex == -1:
				return _crashDoc("Could not find `(` following the `action:` declaration. This needs to be on the same line.", lineNumber)

			actionName = command[comIndex + 1:startIndex].strip()

			# track opening parenthesis to determine the end of the action, set to 1 because the action block is wrapped in parenthesis
			openCount = 1

			firstRow = command[startIndex + 1:].strip()

			# start to build the contents of the action
			actionContents = firstRow + "\n"

			for c in firstRow:
				if c == "(":
					openCount += 1
				elif c == ")":
					openCount -= 1

			# move onto other lines, action block is still open
			while openCount != 0 and lineNumber < len(lines) - 1:
				lineNumber += 1
				newLine = lines[lineNumber].strip()
				for c in newLine:
					if c == "(":
						openCount += 1
					elif c == ")":
						openCount -= 1

					if openCount < 0:
						return _crashDoc("Unexpected `)`.", lineNumber)
				
				actionContents += newLine + "\n"

			if openCount > 0:
				return _crashDoc("Expected ending `)` for action block.", lineNumber)
			elif openCount < 0:
				return _crashDoc("Unexpected `)`.", lineNumber)
			
			actionContents = actionContents[:actionContents.rfind(")")]

			readTo = lineNumber + 1

			res.actions.append(Action(actionName, actionContents))
		elif len(command) > 0:
			return _crashDoc("Could not interpret line format.", lineNumber)
	return res

def digestDeclaration(line: str) -> tuple:
	restype: str = None
	resvalue: str = ""
	for etype in elementTypes:
		if line.startswith(etype):
			restype = etype
			break
	
	if restype == None:
		return None
	else:
		colIndex = line.find(":")
		if colIndex != -1:
			resvalue = line[colIndex + 1:]
		return (restype, resvalue)

def getTop(queue) -> Element:
	return queue[-1] if len(queue) > 0 else None

def digestAttribute(line) -> dict:
	colIndex = line.find(":")
	if colIndex == -1:
		return None
	
	attribName: str = line[:colIndex]
	if len(attribName) == 0:
		return None
	else:
		attribValue: str = line[colIndex + 1:]
		if len(attribValue) == 0:
			return None
		else:
			return { "name": attribName, "value": attribValue.strip() }


import os
import curses
import curses.ascii
import threading
import time

from .adom import Document, Element
from .browser import *
from .render import *
from .util import *
from .window import Window

window = Window(curses.initscr())

browser = Browser("term://welcome")

user_input = None
user_input_thread_handle = None

fps = 30 # 30 frames per second

def setup(screen):
	lifecycle()

def lifecycle():
	global window, user_input_thread_handle

	user_input_thread_handle = threading.Thread(name ='daemon', target=user_input_thread, args=())
	user_input_thread_handle.setDaemon(True)
	user_input_thread_handle.start()

	render()

	while True:
		resized = window.get_resized()
		if resized:
			window.resize()
			continue
		update()
		render()
		if browser.loading:
			browser.start_load()
		window.refresh()
		time.sleep(1.0 / fps)

def user_input_thread():
	global window, user_input
	while True:
		user_input = window.get_input()
		linkIndex = browser.document.find_link(user_input)
		if user_input == ord("`"):
			window.exiting = True
			exit(0)
		elif user_input == 9: # tab:
			browser.document.focus_next()
		elif browser.document.focus != -1:
			char = chr(user_input)
			inputChars = " abcdefghijklmnopqrstuvwxyzABCDEFHIJKLMNOPQRSTUVWXYZ1234567890!@#$%&*()-=_+[]{}\|'\";:.>,</?"

			if browser.document.focus == -2: # URL
				if inputChars.find(char) != -1:
					browser.URL += str(char)
				elif user_input == 127: # backspace
					browser.URL = browser.URL[:-1] if len(browser.URL) > 0 else ""
				elif user_input == 10: # enter
					browser.open_link(browser.URL)
					browser.document.focus = -1
			else: # input element
				focusedElement = browser.document.get_focused_element()
				if inputChars.find(char) != -1:
					focusedElement.value += str(char)
				elif user_input == 127: # backspace
					focusedElement.value = focusedElement.value[:-1] if len(focusedElement.value) > 0 else ""
				elif user_input == 10: # enter
					browser.document.submit(focusedElement)
				elif user_input == 197: # Alt + Q
					browser.document.unfocus()
		elif user_input == 27:
			browser.document.unfocus()
			browser.document.focus = -2
		elif linkIndex != -1:
			exiting = browser.open_link(browser.document.links[linkIndex].URL)

def update():
	global window, user_input
	if window.exiting:
			exit(0)
	user_input = ""

def remove_spacing(string: str):
	return string.replace("\t", "").replace("\n", "")

def render():
	window.start_render(0, 0)

	render_url_length = window.WIDTH - 1
	render_url = expand_len(
		restrict_len(
			browser.URL + ("\N{FULL BLOCK}" if browser.document.focus == -2 else ""),
			render_url_length
		),
		render_url_length
	)

	window.render(render_url, curses.A_UNDERLINE)
	window.render("\N{HEAVY CHECK MARK}" if not browser.loading else "\N{DOTTED CIRCLE}", curses.A_UNDERLINE)
	
	(output, styles) = renderDocument(browser.document, window.WIDTH, window.HEIGHT)
	output = restrict_len(
		remove_spacing(output),
		window.WIDTH * (window.HEIGHT - 1) - 1 # remove last line to account for URL bar, remove last character to stop scroll
	)
	render_pieces = get_pieces(styles, len(output))

	window.start_render(1, 0)
	
	for piece in render_pieces:
		(startPos, endPos, col) = piece
		window.render(output[startPos:endPos], col)
	
	window.disable_cursor()

def get_pieces(styles: list, output_len: int):
	pieces = []
	for i in range(len(styles)):
		current = styles[i]
		follow = None
		if i < len(styles) - 1:
			follow = styles[i + 1]

		startPos = current.start

		endPos = output_len
		if follow != None:
			endPos = follow.start

		col = curses.A_REVERSE

		if current.style == "bold":
			col += curses.A_BOLD
		elif current.style == "underline":
			col += curses.A_UNDERLINE
		pieces.append((startPos, endPos, col))
	return pieces

curses.wrapper(setup)


import sys
import os
import requests as requests

from .term2doc import *

class Browser:
	def __init__(self, initialURL: str):
		self.document = Document(self)
		self.URL = initialURL
		self.loading = True

	def start_load(self):
		if not self.loading:
			return
		self.document = loadFromURL(self.URL, self)
		self.loading = False
		if self.document.hasInputs:
			self.document.focus_next()
	
	# returns if the browser should exit
	def open_link(self, URL: str) -> bool:
		if URL != "term://exit":
			if URL.startswith("term://") and not self.URL.startswith("term://"):
				self.document = Document(self)
				self.document.elements.append(createTextElement("Can not open term:// links. Only Term can open these links."))
			else:
				self.loading = True
				self.URL = URL
			return False
		else:
			return True

"""
term:// -> files in the ./local/ folder of the term project
file:// -> files on the user's computer
http:// & https:// -> files to curl from the internet
"""
def loadFromURL(URL: str, browser: Browser):
	protocol = None
	protocols = ["term://", "http://", "https://"]
	for p in protocols:
		if URL.startswith(p):
			protocol = p
			break
	if protocol == None:
		sys.exit("Error: Could not determine URL protocol `{}`".format(URL))
	elif protocol == "term://":
		file_path = os.path.dirname(os.path.realpath(__file__)) + "/local/" + URL[len("term://"):] + ".term"
		try:
			stream = open(file_path)
			res = stream.read()
			stream.close()
			return term2doc(res, browser)
		except FileNotFoundError:
			return Document(browser).with_message("Could not load `{}`".format(file_path))
	elif protocol == "https://" or protocol == "http://":
		couldOpen = True
		try:
			r = requests.get(URL)
			return term2doc(r.text, browser)
		except:
			return Document(browser).with_message("Could not load URL.")
	



from typing import List
import simpleeval
from urllib.parse import quote

class Action:
	def __init__(self, name, contents):
		self.name = name
		self.code = contents

class Attribute:
	def __init__(self, name, value):
		self.name = name
		self.value = value

class Element:
	def __init__(self, type):
		self.type = type
		self.value = None
		self.attributes = []
		self.children = []
		self.focused = False
	
	def setAttribute(self, name, value):
		self.attributes.append(Attribute(name, value))

	def getAttribute(self, name):
		if(name != ""):
			for item in self.attributes:
				if item.name == name:
					return item.value
			return None
		else:
			return None

	def appendChild(self, child):
		self.children.append(child)

def createTextElement(value):
	t = Element("text")
	t.value = value
	return t
	

class DocumentLink:
	def __init__(self, key: chr, URL: str):
		self.key = key
		self.URL = URL

class Document:
	def __init__(self, browser):
		self.browser = browser
		self.links = []
		self.elements = []
		self.actions = []
		self.focus = -1
		self.hasInputs = False
	
	def with_message(self, message: str):
		self.elements = [createTextElement(message)]
		return self

	def focus_next(self):
		self.focus = self._search_focus(self.focus)

	def get_focused_element(self):
		return self.get_focus_list()[self.focus]

	def unfocus(self):
		focus_list: List[Element] = get_all_elements(self.elements)
		for item in focus_list:
			item.focused = False
		self.focus = -1

	def get_focus_list(self):
		all: List[Element] = get_all_elements(self.elements)
		focusable = ("input",)
		focusList: List[Element] = []
		for element in all:
			if element.type.startswith(focusable):
				focusList.append(element)
		return focusList
	
	def _search_focus(self, start: int):
		all: List[Element] = get_all_elements(self.elements)
		focusable = ("input",)
		focusList: List[Element] = []
		for element in all:
			element.focused = False
			if element.type.startswith(focusable):
				focusList.append(element)
		index = start + 1
		index = index if index < len(focusList) else 0
		if len(focusList) > 0:
			focusList[index].focused = True
			return index
		else:
			return -1

	def add_link(self, key: str, URL: str):
		self.links.append(DocumentLink(ord(key), URL))

	def find_link(self, key: int):
		index = 0
		for link in self.links:
			if link.key == key:
				return index
			index += 1
		return -1

	def find_action(self, name: str) -> Action:
		for a in self.actions:
			if a.name == name:
				return a
		return None

	def submit(self, element: Element):
		actionCall = element.getAttribute("submit")
		if actionCall == None:
			return
		action = self.find_action(actionCall)
		if action == None:
			return
		browser_functions = simpleeval.DEFAULT_FUNCTIONS.copy()
		browser_functions.update(
			visit=(self.browser.open_link),
			encode=(quote)
		)
		simpleeval.simple_eval(action.code, names={"value":element.value}, functions=browser_functions)
		

def get_all_elements(elementList: List[Element]):
	res = []
	for element in elementList:
		res.append(element)
		if len(element.children) > 0:
			res.extend(get_all_elements(element.children))
	return res


class Vec:
	def __init__(self, x: int, y: int):
		self.x = x
		self.y = y
	def add(self, x: int, y: int):
		self.x += x
		self.y += y

def cloneVec(vec: Vec) -> Vec:
	return Vec(vec.x, vec.y)

def addVec(v1: Vec, v2: Vec) -> Vec:
	return Vec(v1.x + v2.x, v1.y + v2.y)

def equalsVec(v1: Vec, v2: Vec) -> bool:
	return v1.x == v2.x and v1.y == v2.y


from .adom import Document, Element
from .vector import *
from .util import *

class OutputStyle:
	def __init__(self, start: Vec, style: str):
		self.start = start
		self.style = style

def renderDocument(document: Document, width: int, height: int):
	res = clearScreen(width, height)
	styles: List[OutputStyle] = [
		OutputStyle(Vec(0, 0), "normal")
	]

	cursor = Vec(0, 0)

	for element in document.elements:
		writeSize = renderElement(element, cursor.x, cursor.y, width, height, res, styles, None)
		cursor.y += writeSize.y

	# combine rows
	out = ""
	for row in res:
		out += row

	# style Vec -> indices
	for style in styles:
		style.start = style.start.x + (style.start.y * width)
	
	return (out, styles)

def clearScreen(width: int, height: int) -> list:
	return [" " * width] * height

allStyles = (
	"bold",
	"underline"
)

def renderElement(element: Element, x: int, y: int, WIDTH: int, HEIGHT: int, res: list, styles: list, parentSize: Vec):
	writeSize = Vec(0, 0)

	if element.type == "cont":
		padding = getPadding(element)
		direction = getDirection(element)
		borderType = element.getAttribute("border")
		paddingSize = Vec(padding['left'] + padding['right'], padding['top'] + padding['bottom'])

		initialOffset = Vec(padding['left'], padding['top'])

		defSize = getDefinedSize(element, WIDTH, HEIGHT)
		hasDefWidth = defSize.x != -1
		hasDefHeight = defSize.y != -1

		if borderType != None:
			initialOffset.add(1, 1)

		offset = cloneVec(initialOffset)

		childrenSize = Vec(0, 0)

		# get children size (uncalculated pc)
		for child in element.children:
			defSingleSize = getDefinedSize(child, WIDTH, HEIGHT)
			otherSingleSize = getElementSize(child)
			singleSize = Vec(
				otherSingleSize.x if defSingleSize.x == -1 else defSingleSize.x,
				otherSingleSize.y if defSingleSize.y == -1 else defSingleSize.y
			)

			if direction == "row":
				childrenSize.x += singleSize.x
				childrenSize.y = max(childrenSize.y, singleSize.y)
			elif direction == "column":
				childrenSize.y += singleSize.y
				childrenSize.x = max(childrenSize.x, singleSize.x)

		# get defined size -> fallback on children collective size
		contSize = Vec(
			defSize.x if hasDefWidth else childrenSize.x + paddingSize.x,
			defSize.y if hasDefHeight else childrenSize.y + paddingSize.y
		)

		# enforce box-sizing
		if borderType != None:
			contSize.add(-2 if hasDefWidth else 0, -2 if hasDefHeight else 0)

		# render children with clamp, calculated pc
		for child in element.children:
			singleSize = renderElement(child, x + offset.x, y + offset.y, WIDTH, HEIGHT, res, styles, contSize)

			if direction == "row":
				offset.x += singleSize.x
			elif direction == "column":
				offset.y += singleSize.y
		
		# render border
		if borderType != None:
			renderBorder(borderType, Vec(x, y), contSize, res)
			writeSize = Vec(contSize.x + 2, contSize.y + 2)
		else:
			writeSize = contSize

	elif element.type == "text" and element.value != "":
		alignOffset = getAlignOffset(element, parentSize)
		writeSize = Vec(len(element.value), 1)
		startPos = x + alignOffset
		textStyle = element.getAttribute("style")
		if textStyle != None and textStyle.startswith(allStyles):
			styles.append(OutputStyle(Vec(startPos, y), textStyle))
			styles.append(OutputStyle(Vec(startPos + len(element.value), y), "normal"))
		newRow = res[y][0:startPos]
		newRow += element.value
		newRow += res[y][startPos + len(element.value):]
		res[y] = newRow
	elif element.type == "link":
		toRender = getLinkText(element)
		toRenderLength = len(toRender)
		alignOffset = getAlignOffset(element, parentSize)
		writeSize = Vec(toRenderLength, 1)
		startPos = x + alignOffset
		textStyle = element.getAttribute("style")
		if textStyle != None and textStyle.startswith(allStyles):
			styles.append(OutputStyle(Vec(startPos, y), textStyle))
			styles.append(OutputStyle(Vec(startPos + toRenderLength, y), "normal"))
		newRow = res[y][0:startPos]
		newRow += toRender
		newRow += res[y][startPos + toRenderLength:]
		res[y] = newRow
	elif element.type == "input":
		defSize = getDefinedSize(element, WIDTH, HEIGHT)
		calcWidth = defSize.x if defSize.x != -1 else 15
		toRender = rcaplen(extlen(element.value + ("\N{FULL BLOCK}" if element.focused else ""), calcWidth), calcWidth)
		toRenderLength = len(toRender)
		alignOffset = getAlignOffset(element, parentSize)
		borderType = "dotted thick" if element.focused else "dotted thin"
		renderBorder(borderType, Vec(x, y), Vec(toRenderLength, 1), res)
		writeSize = Vec(toRenderLength, 3)
		startPos = x + alignOffset + 1
		newRow = res[y + 1][0:startPos]
		newRow += toRender
		newRow += res[y + 1][startPos + toRenderLength:]
		res[y + 1] = newRow

	return writeSize

def getLinkText(element):
	return "[" + element.getAttribute("key") + "] " + element.value 

def getAlignOffset(element: Element, parentSize: Vec) -> int:
	val = element.value
	if element.type == "link":
		val = getLinkText(element)
	align = element.getAttribute("align")
	if align == "center":
		return round(float(parentSize.x) / 2.0) - round(float(len(val)) / 2.0)
	return 0

def getElementSize(element: Element) -> Vec:
	if element.type == "text":
		return Vec(len(element.value), 1)
	elif element.type == "link":
		return Vec(len(getLinkText(element)), 1)
	elif element.type == "input":
		return Vec(len(extlen(element.value, 10)) + 2, 3)
	elif element.type == "cont":
		childrenSize = Vec(0, 0)
		for child in element.children:
			singleSize = getElementSize(child)
			direction = getDirection(element)
			if direction == "row":
				childrenSize.x += singleSize.x
				childrenSize.y = max(childrenSize.y, singleSize.y)
			elif direction == "column":
				childrenSize.y += singleSize.y
				childrenSize.x = max(childrenSize.x, singleSize.x)
		return childrenSize
	return None

def renderBorder(type: str, pos: Vec, size: Vec, res: list):
	screenSize = Vec(len(res[0]), len(res))
	borderCodes = getBorderCodes(type)

	# first row
	firstRow = ""
	if pos.y >= len(res):
			return
	
	# margin
	firstRow += res[pos.y][0:pos.x]
	
	# top left corner
	firstRow += borderCodes["top-left"]
	# top
	firstRow += borderCodes["top"] * size.x
	# top right corner
	firstRow += borderCodes["top-right"]

	firstRow += res[pos.y][pos.x + 1 + size.x + 1:screenSize.x]

	# establish first row
	res[pos.y] = firstRow if len(firstRow) <= screenSize.x else firstRow[:screenSize.x]

	# middle rows
	for y in range(size.y):
		if pos.y + 1 + y >= len(res):
			break
		
		newRow = res[pos.y + 1 + y][0:pos.x]
		newRow += borderCodes["left"]
		newRow += res[pos.y + 1 + y][pos.x + 1:pos.x + 1 + size.x]
		newRow += borderCodes["right"]
		newRow += res[pos.y + 1 + y][pos.x + 1 + size.x + 1:screenSize.x]
		res[pos.y + 1 + y] = newRow if len(newRow) <= screenSize.x else newRow[:screenSize.x]

	# last row
	lastRow = ""
	if pos.y + size.y + 1 >= len(res):
			return

	# margin
	lastRow += res[pos.y + size.y + 1][0:pos.x]
	
	# bottom left corner
	lastRow += borderCodes["bottom-left"]
	# bottom
	lastRow += borderCodes["bottom"] * size.x
	# bottom right corner
	lastRow += borderCodes["bottom-right"]

	lastRow += res[pos.y + size.y + 1][pos.x + 1 + size.x + 1:screenSize.x]
	
	# establish last row
	res[pos.y + size.y + 1] = lastRow if len(lastRow) <= screenSize.x else lastRow[:screenSize.x]

def getDefinedSize(element: Element, WIDTH: int, HEIGHT: int) -> Vec:
	res = Vec(-1, -1)
	defWidth = element.getAttribute("width")
	defHeight = element.getAttribute("height")
	if defWidth != None:
		res.x = parseSize(defWidth, WIDTH)
	if defHeight != None:
		res.y = parseSize(defHeight, HEIGHT)
	return res

def parseSize(size, MAX) -> int:
	if isInt(size):
		return int(size)
	else:
		last = size[-2:]
		start = size[:-2]
		if isInt(start):
			value = int(start)
			if last == "pc":
				return round(float(MAX) / 100.0 * value)
			else:
				return -1
		else:
			return -1

def getPadding(element: Element) -> dict:
	paddingOut = {
		'top': 0,
		'bottom': 0,
		'left': 0,
		'right': 0
	}
	
	padding = element.getAttribute("padding")
	padding_top = element.getAttribute("padding-top")
	padding_bottom = element.getAttribute("padding-bottom")
	padding_left = element.getAttribute("padding-left")
	padding_right = element.getAttribute("padding-right")

	if padding != None:
		paddingVal = int(padding)
		paddingOut = {
			'top': paddingVal,
			'bottom': paddingVal,
			'left': paddingVal,
			'right': paddingVal
		}
	
	if padding_top != None:
		paddingOut['top'] += int(padding_top)
		
	if padding_bottom != None:
		paddingOut['bottom'] += int(padding_bottom)
	
	if padding_left != None:
		paddingOut['left'] += int(padding_left)

	if padding_right != None:
		paddingOut['right'] += int(padding_right)

	return paddingOut

def getDirection(element: Element) -> str:
	dir = element.getAttribute("direction")
	if dir == None:
		return "column"
	else:
		return dir

def getBorderCodes(type: str) -> dict:
	# defaults are for 'line'
	res = {
		"top-left": "\u2554",
		"top-right": "\u2557",
		"bottom-left": "\u255A",
		"bottom-right": "\u255D",
		"top": "\u2550",
		"bottom": "\u2550",
		"left": "\u2551",
		"right": "\u2551",
	}

	if type == "dotted thick":
		res["top-left"] = "\u250F"
		res["top-right"] = "\u2513"
		res["bottom-left"] = "\u2517"
		res["bottom-right"] = "\u251B"
		res["top"] = "\u2505"
		res["bottom"] = "\u2505"
		res["left"] = "\u2507"
		res["right"] = "\u2507"

	if type == "dotted thin":
		res["top-left"] = "\u250C"
		res["top-right"] = "\u2510"
		res["bottom-left"] = "\u2514"
		res["bottom-right"] = "\u2518"
		res["top"] = "\u2504"
		res["bottom"] = "\u2504"
		res["left"] = "\u2506"
		res["right"] = "\u2506"

	return res


import curses

class Window:
	def __init__(self, screen):
		pass
		y, x = screen.getmaxyx()
		self.WIDTH = x
		self.HEIGHT = y
		curses.noecho()
		screen.keypad(True)
		screen.clear()
		self.screen = screen
		self.exiting = False
	def resize(self):
		y, x = self.screen.getmaxyx()
		self.screen.clear()
		curses.resizeterm(y, x)
		self.WIDTH = x
		self.HEIGHT = y
		self.refresh()
	def refresh(self):
		self.screen.refresh()
	def get_resized(self) -> bool:
		return curses.is_term_resized(self.HEIGHT, self.WIDTH)
	def get_input(self) -> chr:
		res = self.screen.getch()
		curses.flushinp()
		return res
	def start_render(self, y: int, x: 0):
		self.screen.move(y, x)
	def render(self, string: str, option):
		self.screen.addstr(string, option)
	def disable_cursor(self):
		curses.curs_set(False)


def isInt(string) -> bool:
	return string.lstrip('-').isdigit()

def restrict_len(string: str, cap: int) -> str:
	return string if len(string) <= cap else string[:cap]

def rcaplen(string: str, cap: int) -> str:
	return string if len(string) <= cap else string[len(string) - cap:]

def expand_len(string: str, cap: int) -> str:
	return string if len(string) >= cap else string + (" " * (cap - len(string)))